import os
from random import randint
from string import ascii_letters

from ._exceptions import DeleteError, PermissionError, out_of_space
from ._abstract import AbstractVlermv
from ._exceptions import OpenError

def _get_fn(fn, mode, load):
    '''
    Load a contents, checking that the file was not modified during the read.
    '''
    try:
        mtime_before = os.path.getmtime(fn)
    except OSError:
        mtime_before = None

    try:
        with open(fn, mode) as fp:
            item = load(fp)
    except OpenError:
        raise
    else:
        mtime_after = os.path.getmtime(fn)
        if mtime_before in {None, mtime_after}:
            return item
        else:
            raise EnvironmentError('File was edited during read: %s' % fn)


def _random_file_name():
    n = len(ascii_letters) - 1
    return ''.join(ascii_letters[randint(0, n)] for _ in range(10))

def mktemp(tempdir, filename = _random_file_name):
    try:
        os.makedirs(tempdir, exist_ok = True)
    except TypeError:
        # Python 2
        if not os.path.isdir(tempdir):
            os.makedirs(tempdir)
    return os.path.join(tempdir, filename())

def _reversed_directories(outer, inner):
    while outer != inner:
        yield inner
        inner = os.path.dirname(inner)

class Vlermv(AbstractVlermv):
    '''
    A :py:class:`dict` API to a filesystem
    '''

    #: Should the cache directory be created when a Vlermv is initialized?
    #: This is is mostly relevant for testing.
    _mkdir = True

    def __init__(self, *directory, tempdir = '.tmp', **kwargs):
        '''
        :param str directory: Top-level directory of the vlermv
        :param serializer: A thing with dump and load functions for
            serializing and deserializing Python objects,
            like :py:mod:`json`, :py:mod:`yaml`, or
            anything in :py:mod:`vlermv.serializers`
        :type serializer: :py:mod:`serializer <vlermv.serializers>`
        :param key_transformer: A thing with to_path and from_path functions
            for transforming keys to file paths and back.
            Several are available in :py:mod:`vlermvtransformers`.
        :type key_transformer: :py:mod:`key_transformer <vlermvtransformers>`
        :param bool mutable: Whether values can be updated and deleted
        :param str tempdir: Subdirectory inside of base_directory to use for temporary files

        These are mostly relevant for initialization via :py:func:`vlermv.cache`.

        :param bool appendable: Whether new values can be added to the Vlermv
            (Set this to False to ensure that the decorated function is never
            run and that the all results are cached; this is useful for reviewing
            old data in a read-only mode.)
        :param bool cache_exceptions: If the decorated function raises
            an exception, should the failure and exception be cached?
            The exception is raised either way.
        :raises TypeError: If cache_exceptions is True but the serializer
            can't cache exceptions
        '''
        super(Vlermv, self).__init__(**kwargs)
        self.base_directory = os.path.expanduser(os.path.join(*directory))
        self.tempdir = os.path.join(self.base_directory, tempdir)
        if self._mkdir:
            os.makedirs(self.tempdir, exist_ok = True)

    def __repr__(self):
        return 'Vlermv(%s)' % repr(self.base_directory)

    def __setitem__(self, index, obj):
        super(Vlermv, self).__setitem__(index, obj)
        fn = self.filename(index)
        os.makedirs(os.path.dirname(fn), exist_ok = True)
        exists = os.path.exists(fn)
        if (not self.mutable) and exists:
            raise PermissionError('This warehouse is immutable, and %s already exists.' % fn)
        elif (not self.appendable) and (not exists):
            raise PermissionError('This warehouse not appendable, and %s does not exist.' % fn)
        else:
            tmp = mktemp(self.tempdir)
            with open(tmp, 'w+' + self._b()) as fp:
                try:
                    self.serializer.dump(obj, fp)
                except Exception as e:
                    if out_of_space(e):
                        fp.close()
                        os.remove(tmp)
                        raise BufferError('Out of space')
                    else:
                        raise
            os.rename(tmp, fn)

    def __getitem__(self, index):
        try:
            return _get_fn(self.filename(index), 'r+' + self._b(), self.serializer.load)
        except OpenError:
            raise KeyError(index)

    def __delitem__(self, index):
        super(Vlermv, self).__delitem__(index)
        fn = self.filename(index)
        try:
            os.remove(fn)
        except DeleteError as e:
            raise KeyError(*e.args)
        else:
            for fn in _reversed_directories(self.base_directory, os.path.dirname(fn)):
                if os.listdir(fn) == []:
                    os.rmdir(fn)
                else:
                    break

    def __len__(self):
        length = 0
        for dirpath, _, filenames in os.walk(self.base_directory):
            for filename in filenames:
                length += 1
        return length

    def keys(self):
        for dirpath, _, filenames in os.walk(self.base_directory):
            if dirpath != os.path.join(self.base_directory, self.tempdir):
                for filename in filenames:
                    path = self.from_filename(os.path.abspath(os.path.join(dirpath, filename)))
                    if path != None:
                        yield path



import tempfile, socket

from ._abstract import AbstractVlermv
from ._safe_buckets import SafeBuckets

class S3Vlermv(AbstractVlermv):
    buckets = SafeBuckets()

    def __init__(self, bucketname, *path, bucket = None, **kwargs):
        super(S3Vlermv, self).__init__(**kwargs)
        if bucket:
            self.bucket = bucket
        else:
            self.bucket = self.buckets[bucketname]

        self.base_directory = '/'.join(path)
        if self.base_directory != '':
            self.base_directory += '/'

    def __repr__(self):
        return 'S3Vlermv(%s/%s)' % (self.bucket.name, self.base_directory)

    def __setitem__(self, index, obj):
        super(S3Vlermv, self).__setitem__(index, obj)
        keyname = self.filename(index)
        key = self.bucket.new_key(keyname)
        with tempfile.NamedTemporaryFile('w+' + self._b()) as tmp:
            self.serializer.dump(obj, tmp.file)
            tmp.file.close()
            key.set_contents_from_filename(tmp.name, replace = True)

    def __contains__(self, index):
        keyname = self.filename(index)
        return self.bucket.get_key(keyname) != None

    class Timeout(socket.timeout):
        pass

    def __getitem__(self, index):
        keyname = self.filename(index)
        key = self.bucket.get_key(keyname)
        if key:
            with tempfile.NamedTemporaryFile('w+' + self._b()) as tmp:
                try:
                    key.get_contents_to_filename(tmp.name)
                except socket.timeout:
                    raise self.__class__.Timeout('Timeout when reading from S3')

                tmp.file.seek(0)

                try:
                    value = self.serializer.load(tmp.file)
                except FileNotFoundError:
                    raise self.__class__.Timeout('Timeout when reading from S3')

            return value
        else:
            raise KeyError(keyname)

    def keys(self):
        for k in self.bucket.list(prefix = self.base_directory):
            index = self.from_filename(k.name)
            if index != None:
                yield index

    def __delitem__(self, index):
        super(S3Vlermv, self).__delitem__(index)
        keyname = self.filename(index)
        self.bucket.delete_key(keyname)

    def __len__(self):
        return sum(1 for _ in self.keys())



import os

class delimit:
    def __init__(self, delimiter):
        self.delimiter = delimiter

    def to_path(self, key):
        return tuple(key.split(self.delimiter))

    def from_path(self, obj):
        return self.delimiter.join(obj)

slash = delimit('/')
backslash = delimit('\\')



from base64 import b64encode, b64decode

def to_path(path):
    return (b64encode(path.encode('utf-8')).decode('ascii'),)

def from_path(key):
    if len(key) != 1:
        raise ValueError('Key must have exactly one element.')
    return b64decode(key[0]).decode('ascii')



class _meta_xml:
    @staticmethod
    def dump(obj, fp):
        fp.write(self.module.tostring(obj))

    @staticmethod
    def load(fp):
        return self.module.fromstring(fp.read())

    cache_exceptions = False
    binary_mode = False

try:
    import lxml
except ImportError:
    pass
else:
    import lxml.html, lxml.etree
    class html(_meta_xml):
        module = lxml.html
    class xml(_meta_xml):
        module = lxml.etree



ERROR = TypeError('key must be a tuple of strings.')
def to_path(key):
    if not isinstance(key, tuple):
        raise ERROR
    for x in key:
        if not isinstance(x, str):
            raise ERROR
    return key

def from_path(x):
    return x



import os, posixpath

def split(path):
    'Split a path into a tuple of all directories and then the final directory/file.'
    if path == '':
        raise ValueError('Path may not be empty.')

    dn, fn = os.path.split(path)
    if fn == '':
        return (dn,)
    elif dn == '':
        return (fn,)
    else:
        return split(dn) + (fn,)

def method_or_name(namespace, x):
    '''
    If x is a ``str``, get ``namespace.x``.

    Otherwise, simply return ``x``.
    '''
    if not isinstance(x, str):
        return x

    if hasattr(namespace, x):
        return getattr(namespace, x)

    attrs = [y for y in dir(namespace) if not y.startswith('_')]
    msg = '''"%s" is not available in %s.
These attributes are available: %s.''' % (x, namespace.__name__, attrs)
    raise AttributeError(msg)

def safe_path(unsafe_path, not_allowed = {'', '.'}):
    unsafe_str = posixpath.join('', *unsafe_path)
    rough_tuple = posixpath.join('/', posixpath.normpath(unsafe_str)).split('/')[1:]
    safe_tuple = tuple(x for x in rough_tuple if x not in not_allowed)

    if len(safe_tuple) == 0:
        raise ValueError('Path must not be empty.')
    elif safe_tuple[0] == '..':
        raise ValueError('Relative paths above the starting directory (./..) are not allowed.')
    else:
        return safe_tuple



import os

error_msg = '''The default transformer cannot handle slashes (subdirectories);
try another transformer in vlermv.transformers.'''

def to_path(key):
    if '/' in key or '\\' in key or os.path.sep in key:
        raise ValueError(error_msg)
    return (key,)

def from_path(path):
    if len(path) != 1:
        raise ValueError(error_msg)
    return path[0]



import os, re

from ._exceptions import PermissionError
from .serializers import pickle
from .transformers import magic
from ._util import safe_path

import logging

logger = logging.getLogger(__name__)

class AbstractVlermv:
    '''
    A :py:class:`dict` API to various things
    '''

    @classmethod
    def memoize(Class, *args, **kwargs):
        '''
        Memoize/record a function inside this vlermv. ::

            @Vlermv.cache('~/.http')
            def get(url):
                return requests.get(url, auth = ('username', 'password'))

        The args and kwargs get passed to the Vlermv with some slight changes.
        Here are the changes.

        First, the default ``key_transformer`` is the tuple key_transformer
        rather than the simple key_transformer.

        Second, it is valid for cache to be called without arguments.
        Vlermv would ordinarily fail if no arguments were passed to it.
        If you pass no arguments to cache, the Vlermv directory argument
        (the one required argument) will be set to the name of the function.

        Third, you are more likely to use the ``cache_exceptions`` keyword
        argument; see :py:class:`~vlermv.Vlermv` for documentation on that.
        '''
        def decorator(func):
            if len(args) == 0:
                if hasattr(func, '__name__'): 
                    _args = (func.__name__,)
                else:
                    raise ValueError('You must specify the location to store the vlermv.')
            else:
                _args = args
            v = Class(*_args, **kwargs)
            v.func = func
            return v
        return decorator

    serializer = pickle
    key_transformer = magic
    appendable = True
    mutable = True
    tempdir = '.tmp'
    cache_exceptions = False
    extension = ''
    base_directory = ''

    def __init__(self, **kwargs):
        '''
        :param str base_directory: Top-level directory of the vlermv
        :param serializer: A thing with dump and load functions for
            serializing and deserializing Python objects,
            like :py:mod:`json`, :py:mod:`yaml`, or
            anything in :py:mod:`vlermv.serializers`
        :type serializer: :py:mod:`serializer <vlermv.serializers>`
        :param key_transformer: A thing with to_path and from_path functions
            for transforming keys to file paths and back.
            Several are available in :py:mod:`vlermvtransformers`.
        :type key_transformer: :py:mod:`key_transformer <vlermvtransformers>`
        :param bool mutable: Whether values can be updated and deleted
        :param str tempdir: Subdirectory inside of base_directory to use for temporary files

        These are mostly relevant for initialization via :py:func:`vlermv.cache`.

        :param bool appendable: Whether new values can be added to the Vlermv
            (Set this to False to ensure that the decorated function is never
            run and that the all results are cached; this is useful for reviewing
            old data in a read-only mode.)
        :param bool cache_exceptions: If the decorated function raises
            an exception, should the failure and exception be cached?
            The exception is raised either way.
        :raises TypeError: If cache_exceptions is True but the serializer
            can't cache exceptions
        '''
        for key in ['serializer', 'appendable', 'mutable', 'base_directory',
                    'key_transformer', 'cache_exceptions', 'extension']:
            setattr(self, key, kwargs.get(key, getattr(self.__class__, key)))

        if self.cache_exceptions and not getattr(self.serializer, 'cache_exceptions', True):
            msg = 'Serializer %s cannot cache exceptions.'
            raise TypeError(msg % repr(self.serializer))

        self.binary_mode = getattr(self.serializer, 'binary_mode', False)
        self.func = None

    def __call__(self, *args, **kwargs):
        if self.func == None:
            msg = 'Set %s.func to something if you want to call %s.'
            raise NotImplementedError(msg % (self, self))
        
        if not hasattr(self.func, '__call__'):
            raise AttributeError('%s.func must be callable.' % self.__class__.__name__)

        if args in self:
            output = self[args]
        else:
            try:
                result = self.func(*args, **kwargs)
            except Exception as error:
                signature = self.__class__.__name__, getattr(self.func, '__name__', str(self.func)), args, kwargs
                msg = 'Exception in %s calling this memoized function:\n%s(*%s, **%s)' % signature
                logger.error(msg, exc_info = False)
                if self.cache_exceptions:
                    output = error, None
                else:
                    raise error
            else:
                if self.cache_exceptions:
                    output = None, result
                else:
                    output = result
            self[args] = output

        if self.cache_exceptions:
            if len(output) != 2:
                msg = '''Deserializer returned %d elements,
but it is supposed to return only two (exception, object).
Perhaps the serializer doesn't implement exception caching properly?'''
                raise TypeError(msg % len(output))

            error, result = output
            if error != None and result != None:
                raise TypeError('''The exception or the object (or both) must be None.
There's probably a problem with the serializer.''')

            if error:
                raise error
        else:
            result = output

        return result

    def filename(self, index):
        '''
        Get the absolute filename corresponding to a key; run the
        key_transformer on the key and do a few other small things.

        :raises TypeError: if the key_transformer returns something other than
            a :py:class:`tuple` of :py:class:`strings <str>`
        :raises KeyError: if the key_transformer returns an empty path
        :returns: the filename
        :rtype: str
        '''
        subpath = self.key_transformer.to_path(index)
        if not isinstance(subpath, tuple):
            msg = 'subpath is a %s, but it should be a tuple.'
            raise TypeError(msg % type(subpath).__name__)
        elif len(subpath) == 0:
            raise KeyError('You specified an empty key.')
        elif not all(isinstance(x, str) for x in subpath):
            msg = 'Elements of subpath should all be str; here is subpath:\n%s' % repr(subpath)
            raise TypeError(msg)
        return os.path.join(self.base_directory, *safe_path(subpath)) + self.extension

    def from_filename(self, filename):
        '''
        Convert an absolute filename into key.
        '''
        i = len(self.base_directory)
        if filename[:i] != self.base_directory:
            raise ValueError('Filename needs to start with "%s";\nyou passed "%s".' % (self.base_directory, filename))

        if filename.endswith(self.extension):
            if len(self.extension) > 0:
                j = -len(self.extension)
            else:
                j = None
            return self.key_transformer.from_path(tuple(filename[i:j].strip('/').split('/')))

    def __iter__(self):
        return (k for k in self.keys())

    def _b(self):
        return 'b' if self.binary_mode else ''

    def __delitem__(self, index):
        if not self.mutable:
            raise PermissionError('This vlermv is immutable, so you can\'t delete things.')

    def __contains__(self, index):
        fn = self.filename(index)
        return os.path.isfile(fn)

    def values(self):
        for key, value in self.items():
            yield value

    def update(self, d):
        generator = d.items() if hasattr(d, 'items') else d
        for k, v in generator:
            self[k] = v

    def get(self, index, default = None):
        if index in self:
            return self[index]
        else:
            return default

    def items(self):
        for key in self.keys():
            yield key, self[key]

    def __setitem__(self, index, obj):
        if (not self.mutable) and (index in self):
            raise PermissionError('This vlermv is not mutable, so you can\'t edit things.')
        if (not self.appendable) and (index not in self):
            raise PermissionError('This vlermv is not appendable, so you can\'t append new things.')

    def __getitem__(self, index):
        raise NotImplementedError

    def keys(self, **kwargs):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError



import posixpath

error_msg = '''The index must be a string.'''

def to_path(key):
    if isinstance(key, tuple) and len(key) == 1:
        key = key[0]

    if hasattr(key, 'strip') and hasattr(key, 'split'):
        return tuple(key.strip('/').split('/'))
    else:
        raise TypeError('Key must be string-like or a tuple of length one.')

def from_path(path):
    return posixpath.join(*path)



'''
Come up with a reasonable file name from a wide range of possible
input formats. (Multiple inputs can map to the same file name.)
'''
import os
try:
    from urllib.parse import urlsplit
except ImportError:
    import urllib2
    urlsplit = urllib2.urlparse.urlsplit
import warnings
import itertools
import datetime

# For Python 2 compatibility
try:
    basestring
except NameError:
    basestring = str

def cli():
    import sys
    if len(sys.argv) > 1:
        sys.stdout.write(os.path.join(*to_path(sys.argv[1:])) + '\n')
    else:
        sys.stderr.write('You must pass at least one argument.\n')
        sys.exit(1)

def from_path(obj):
    return obj

def to_path(index):
    if not safe_type(index):
        warnings.warn(UserWarning('You should pass an object with a deterministic order. (probably not a %s)' % type(index).__name__))

    return tuple(filter(lambda x: x!= '', replace_special(parse(index))))

def parse(index):
    for theclass in [basestring, datetime.date, datetime.datetime, int, type(None)]:
        if isinstance(index, theclass):
            yield from parse_partial(index) # tuple
            break
    else:
        yield from itertools.chain(*map(parse, index)) # iterable

_special = {'.': '\\.', '..': '\\..', '.tmp': '\\.tmp'}
def replace_special(path):
    for item in path:
        if item in _special:
            yield _special[item]
        else:
            yield item

def parse_partial(item):
    if isinstance(item, basestring):
        func = parse_partial_text
    elif isinstance(item, int):
        func = lambda x: (str(x),)
    elif isinstance(item, datetime.date) or isinstance(item, datetime.datetime):
        func = parse_partial_date
    elif item == None:
        func = lambda _: ('',)
    else:
        raise ValueError('item must be string, datetime.date, datetime.datetime or int')
    return tuple(func(item))
#   yield from func(item)

def parse_partial_text(item):
    for a in parse_partial_url(item):
        for b in a.split('\\'):
            if b != '':
                yield b

def parse_partial_url(item):
    url = urlsplit(item)
    path = []

    if url.scheme:
        path.append(url.scheme)

    if url.netloc:
        path.append(url.netloc)

    if url.path:
        for y in filter(None, url.path.split('/')):
            path.append(y)

    if len(path) == 0:
        path = ['']

    if url.query:
        path[-1] += '?' + url.query

    if url.fragment:
        path[-1] += '#' + url.fragment

    return tuple(path)

def parse_partial_date(item):
    date = ('%04d' % item.year, '%02d' % item.month, '%02d' % item.day)
    if hasattr(item, 'minute'):
        time = ('%02d' % item.hour, '%02d' % item.minute)
    else:
        time = tuple()
    return date + time

def safe_type(index):
    for bad in [set,dict]:
        if isinstance(index, bad):
            return False
    else:
        return True



import mmap
import os

class _identity:
    'Dump and load things that are already serialized.'
    @staticmethod
    def dump(obj, fp):
        fp.write(obj)

    @staticmethod
    def load(fp):
        return fp.read()

    cache_exceptions = False

class identity_str(_identity):
    'Dump and load raw strings.'
    binary_mode = False

class identity_bytes(_identity):
    'Dump and load raw bytes.'
    binary_mode = True

class identity_mmap_bytes(_identity):
    'Dump and load raw bytes, loading with a memory-mapped file.'
    binary_mode = True

    @staticmethod
    def load(fp):
        if os.stat(fp.name).st_size > 0:
            return mmap.mmap(fp.fileno(), 0, prot = mmap.PROT_READ).read()
        else:
            return b''

class identity_mmap_str(_identity):
    'Dump and load raw strings, loading with a memory-mapped file.'
    binary_mode = True

    @staticmethod
    def dump(obj, fp):
        fp.write(obj.encode('utf-8'))

    @staticmethod
    def load(fp):
        return identity_mmap_bytes.load(fp).decode('utf-8')



try:
    FileNotFoundError
except NameError:
    OpenError = IOError
else:
    OpenError = FileNotFoundError

try:
    PermissionError
except NameError:
    class PermissionError(EnvironmentError):
        pass
else:
    PermissionError = PermissionError

try:
    FileNotFoundError
except NameError:
    DeleteError = OSError
else:
    DeleteError = FileNotFoundError

def out_of_space(exception):
    return isinstance(exception, IOError) and exception.args == (28, 'No space left on device')



from distutils.core import setup

setup(name='vlermv',
      author='Thomas Levine',
      author_email='_@thomaslevine.com',
      description='Easily dump python objects to files, and then load them back.',
      url='https://thomaslevine.com/!/vlermv/',
      packages=['vlermv', 'vlermv.serializers', 'vlermv.transformers'],
      install_requires = [
          'thready>=0.1.4',
          'boto>=2.38.0',
      ],
      tests_require = [
          'pytest>=2.6.4', 'testfixtures>=4.1.2',
      ],
      version='1.4.0',
      license='AGPL',
      entry_points = {'console_scripts': ['dadaname = vlermv.transformers.magic:cli']},
)



from time import sleep
import threading, queue

from boto import connect_s3

def create_buckets(create_bucket, log, state):
    while True:
        bucketname = log.get()
        if create_bucket == None:
            create_bucket = connect_s3().create_bucket
        state[bucketname] = create_bucket(bucketname)

class SafeBuckets:
    '''
    Thread-safely create a bucket.
    '''
    def __init__(self, create_bucket = None):
        self.log = queue.Queue()
        self.state = {}
        creator = threading.Thread(target = create_buckets,
           args = (create_bucket, self.log, self.state), daemon = True)
        creator.start()

    def __getitem__(self, k):
        self.log.put(k)
        while k not in self.state:
            sleep(0)
        return self.state[k]
